/*
 * actions.c : Widget action procedures
 *
 * This file defines the global actionTable used in XtAppAddActions() and
 * defines the action procedures, except those used on the settingsPanel
 * which are imported. Eventually there may be more actions, and they'll
 * be put here.
 *
 * Thanks to:
 * George Ferguson, ferguson@cs.rochester.edu, 12 Sep 1991.
 *
 */

#include "geotraceman.h"
#include "items.h"

XtAppContext appContext;

/*
 * Functions defined here
 */
extern void addActions();
extern void blank();
extern void Contin_Rotate();
extern Boolean contin_rot();
static void quitAction(); 
static void trace_startAction(), trace_stopAction();
static void trace_removeAction();
static void map_rotateAction(), map_zoomAction(), map_includeAction(), map_resetAction();
static void map_zoomresetAction(), map_zoomcancelAction();
static void trace_changelabelAction();
static void trace_loadAction(), trace_saveAction(), map_3DAction();
static void map_controtAction();
static void postscript_dumpAction();

/*
 * Data defined here:
 */
static XtActionsRec Action_Table[] = {
	{"quit", quitAction},
	{"trace_start", trace_startAction},
	{"trace_stop", trace_stopAction},
	{"trace_remove", trace_removeAction},
	{"map_zoomreset", map_zoomresetAction},
	{"map_zoomcancel", map_zoomcancelAction},
	{"map_rotate", map_rotateAction},
	{"map_zoom", map_zoomAction},
	{"map_include", map_includeAction},
	{"map_reset", map_resetAction},
	{"map_3D", map_3DAction},
	{"map_controt", map_controtAction},
	{"trace_changelabel", trace_changelabelAction},
	{"trace_load", trace_loadAction},
	{"trace_save", trace_saveAction},
	{"postscript_dump", postscript_dumpAction},
	{"Ok",Ok},
	NULL
};

/*	-	-	-	-	-	-	-	-	*/

/*
Registers the actions with the server
*/
void
add_actions(app_context)
        XtAppContext app_context;
{
        XtAppAddActions(app_context,Action_Table,XtNumber(Action_Table));
}



void blank(widget, client_data, call_data)
Widget widget;
XtPointer client_data, call_data;
{
}

ACTION (quitAction)
{
    XtDestroyApplicationContext (appContext);
    exit (0);
}

ACTION (trace_startAction)
{
	Widget widget;
	XtPointer client_data, call_data;
	widget = toplevel;
	/* Test to see if if the user is root, if not exit */
	if (geteuid()) {
		DisplayMessage(widget, "Start Trace", "Must be root to start a trace.");
	}
	else {
		if ((ts.numtrace < Max_Trace) && (ts.tracing == False)) {
			GetStringDialog(widget, "Trace to...", "Enter Destination Host IP Name",StartTrace);
		}
		else {
			DisplayMessage(widget, "Start Trace?", "A trace already exists!");	
		}
	}
}

ACTION (trace_stopAction)
{
	Widget widget;
	Widget tmpWidget;
	XtPointer client_data, call_data;
	widget = toplevel;

	/* kill the child process that is performing the trace */
	if (ts.tracepid != -1)
	{
		kill(ts.tracepid, SIGKILL);
		ts.tracing = False;
		ts.numtrace = 0;
		ts.traces = NULL;
		close(ts.tracesd);
	        if ((tmpWidget = XtNameToWidget(toplevel, "*tstatusnLabel")) == NULL)
       	         	fail0("No such widget tstatusnLabel");
       	 	setLabel(tmpWidget , "waiting");
	}	
}

ACTION (map_rotateAction)
{
	long rotation = (long) (((float) ms.max_range) * ms.rotation_fraction);

	if ((int) *num_params != 1) {
		fprintf(stderr,"GTM: Error wrong number of parameters to map_rotateAction\n");
		exit(1);
	}

	if (strcmp("l", params[0]) == 0) {
		rotation = -rotation;
		update_x_offset(rotation);
		update_region_delimiters();
		update_display(); 
	}
	else if (strcmp("r", params[0]) == 0) {
		update_x_offset(rotation);
		update_region_delimiters();
		update_display(); 
	}

/*
	if ((ms.includes & INCL_3D) == INCL_3D) {
		if (strcmp("u", params[0]) == 0 ) {
			update_y_offset(rotation);
			update_region_delimiters();
			update_display(); 
		}
		else if (strcmp("d", params[0]) == 0) {
			rotation = -rotation;
			update_y_offset(rotation);
			update_region_delimiters();
			update_display(); 
		}
	}
*/
}

ACTION (map_zoomAction)
{
	Widget tmpWidget;

        if ((tmpWidget = XtNameToWidget(toplevel, "*mstatusnLabel")) == NULL)
                fail0("No such widget mstatusnLabel");
        setLabel(tmpWidget , "zooming");
}


ACTION (map_includeAction)
{
	if ((int) *num_params != 1) {
		fprintf(stderr,"GTM: Error wrong number of parameters to map_includeAction\n");
		exit(1);
	}

	if (strcmp("Rivers", params[0]) == 0) {
		if ((ms.includes & INCL_RIV) != INCL_RIV) {
			setTick(ITEM_INCLUDE_RIVS);
			ms.includes |= INCL_RIV;
			color(Map_Color_Riv);
			display_regions(riv_map);
			swapbuffers();
		}
		else {
			unsetTick(ITEM_INCLUDE_RIVS);
			ms.includes &= ~INCL_RIV;
			update_display();
		}
	}

	else if(strcmp("Political_Boundarys", params[0]) == 0) {
		if ((ms.includes & INCL_PDY) != INCL_PDY) {
			setTick(ITEM_INCLUDE_PBYS);
			ms.includes |= INCL_PDY;
			color(Map_Color_Pby);
			display_regions(pdy_map);
			swapbuffers();
		}
		else {
			unsetTick(ITEM_INCLUDE_PBYS);
			ms.includes &= ~INCL_PDY;
			update_display();
		}
	}

	else if(strcmp("Cities", params[0]) == 0) {
		if ((ms.includes & INCL_CIT) != INCL_CIT) {
			setTick(ITEM_INCLUDE_CITS);
			ms.includes |= INCL_CIT;
			color(Map_Color_Cit);
			cities_display(); 
		}
		else {
			unsetTick(ITEM_INCLUDE_CITS);
			ms.includes &= ~INCL_CIT;
			update_display();	
		}
	}
}

ACTION (trace_changelabelAction)
{
	Widget widget = toplevel;

	if ((int) *num_params != 1) {
		fprintf(stderr,"GTM: Error wrong number of parameters to trace_changelabelAction\n");
		exit(1);
	}
	/*GetStringDialog(widget, "Which Trace ?", "Enter Trace Number (0 for all traces", blank);*/

	if (ts.numtrace == 0) {
		DisplayMessage(widget, "What Trace?", "No current trace to change label of");
	}

	else {
		if(strcmp("Position", params[0]) == 0) {
			if (ts.traces->label_position == True) {
				unsetTick(ITEM_LABEL_POSITION);
				ts.traces->label_position = False;
			}
			else {
				setTick(ITEM_LABEL_POSITION);
				ts.traces->label_position = True;
			}
			update_display();
		}
		else if (strcmp("IP_Name", params[0]) == 0) {
			if (ts.traces->label_ipname == True) {
				unsetTick(ITEM_LABEL_IPNAME);
				ts.traces->label_ipname = False;
			}
			else {
				setTick(ITEM_LABEL_IPNAME);
				ts.traces->label_ipname = True;
			}
			update_display();
		}

		else if(strcmp("IP_Num", params[0]) == 0) {
			if (ts.traces->label_ipnum == True) {
				unsetTick(ITEM_LABEL_IPNUM);
				ts.traces->label_ipnum = False;
			}
			else {
				setTick(ITEM_LABEL_IPNUM);
				ts.traces->label_ipnum = True;
			}
			update_display();
		}
	}
}

ACTION(map_resetAction) {
	Widget tmpWidget;

	init_map_state();
	if (ms.detail != Initial_Map_Detail)
		swap_maps(Initial_Map_Detail);

        if ((tmpWidget = XtNameToWidget(toplevel, "*zoomnLabel")) == NULL)
                fail0("No such widget zoomnLabel");
        setLabel(tmpWidget , "1");

        if ((tmpWidget = XtNameToWidget(toplevel, "*currentnzoomLabel")) == NULL)
                fail0("No such widget currentnzoomLabel");
        setLabel(tmpWidget , "1:1");

	update_display();
}

ACTION(trace_removeAction) {
	Widget tmpWidget;

	if (!ts.tracing) {
		ts.traces=(struct Trace_Struct *) NULL;
        	ts.numtrace=0;

       		if ((tmpWidget = XtNameToWidget(toplevel, "*tstatusnLabel")) == NULL)
               		fail0("No such widget tstatusnLabel");
	        setLabel(tmpWidget , "waiting");
       		if ((tmpWidget = XtNameToWidget(toplevel, "*lasthostnLabel")) == NULL)
               		fail0("No such widget lasthostnLabel");
	        setLabel(tmpWidget , "0 ms");
       		if ((tmpWidget = XtNameToWidget(toplevel, "*numhostsnLabel")) == NULL)
               		fail0("No such widget numhostsnLabel");
	        setLabel(tmpWidget , "0");
       		if ((tmpWidget = XtNameToWidget(toplevel, "*totaltimenLabel")) == NULL)
               		fail0("No such widget totaltimenLabel");
	        setLabel(tmpWidget , "0 ms");

        	update_display();
		cities_display();
	}
}

ACTION(map_zoomresetAction) {
	Widget tmpWidget;

	init_map_state();

        if ((tmpWidget = XtNameToWidget(toplevel, "*zoomnLabel")) == NULL)
                fail0("No such widget zoomnLabel");
        setLabel(tmpWidget , "1");

        if ((tmpWidget = XtNameToWidget(toplevel, "*currentnzoomLabel")) == NULL)
                fail0("No such widget currentnzoomLabel");
        setLabel(tmpWidget , "1:1");

	UndoMap(ZOOM_CLEAR);
        update_display();
}

ACTION(map_zoomcancelAction) {
	Widget tmpWidget;

	if (UndoMap(ZOOM_UNDO) == True) {
		if ((ms.includes & INCL_RIV) == INCL_RIV)
			setTick(ITEM_INCLUDE_RIVS);
		else
			unsetTick(ITEM_INCLUDE_RIVS);
		if ((ms.includes & INCL_PDY) == INCL_PDY)
			setTick(ITEM_INCLUDE_PBYS);
		else
			unsetTick(ITEM_INCLUDE_PBYS);
		if ((ms.includes & INCL_CIT) == INCL_CIT)
			setTick(ITEM_INCLUDE_CITS);
		else
			unsetTick(ITEM_INCLUDE_CITS);
		if ((ms.includes & INCL_3D) == INCL_3D)
			setTick(ITEM_INCLUDE_3D);
		else
			unsetTick(ITEM_INCLUDE_3D);
	}
	swap_maps();
	update_display();
}


ACTION(trace_loadAction) {
	if (ts.tracing)
		DisplayMessage(toplevel, "Error", "A trace is currently in progress");
	else
		DisplayMessage(toplevel, "Information", "Not yet implemented");
}

ACTION(trace_saveAction) {
	if (ts.tracing)
		DisplayMessage(toplevel, "Error", "A trace is currently in progress");
	else
		DisplayMessage(toplevel, "Information", "Not yet implemented");
}

ACTION(map_3DAction) {
	if ((ms.includes & INCL_3D) != INCL_3D) {
		setTick(ITEM_INCLUDE_3D);
		ms.includes |= INCL_3D;
	}
	else {
		unsetTick(ITEM_INCLUDE_3D);
		ms.includes &= ~INCL_3D;
	}
	update_display();
}


ACTION (map_controtAction) {

	if (ms.rotating == False)
	{
		GetStringDialog(w, "Continuous Rotation", "Direction to rotate ('right' or 'left')", Contin_Rotate);
		ms.rotating = True;
        	XtAppAddWorkProc(appContext, (XtWorkProc) contin_rot, (caddr_t) NULL);
	}
	else
	{
		ms.rotating = False;
		ms.rot_dir = '?';
	}
}

ACTION (postscript_dumpAction) {
	page_flag = 0;
	page_begin = 1;	

	ps_color = 3;
	ps_dump_regions(cil_map);

	if ((ms.includes & INCL_PDY) == INCL_PDY) {
		ps_color = 0;
		ps_dump_regions(pdy_map);
	}

	if ((ms.includes & INCL_RIV) == INCL_RIV) {
		ps_color = 1;
		ps_dump_regions(riv_map);
	}

	if ((ms.includes & INCL_CIT) == INCL_CIT) {
		ps_color = 2;
		cities_display_ps();
	}	

	display_traces_ps();
	page_flag = 1;	
	ps_dump_regions(cil_map);
}	
	
	
