#include "geotraceman.h"

#define SQR(x) (x * x)

/***
Functions to do with drawing the map to the display 
***/

Boolean convert_values();

/*
Determines the new x offsets from the value passed to the function that the map should be rotated,
either positive or negative
*/
void
update_x_offset(amount)
	long amount;
{
	long tmp_x_offset;
	int tmp_reg_offset;

	if (amount >= 0) {
		tmp_x_offset = (((long) ms.x_offset) + amount);
		if ( tmp_x_offset > (ms.region_range - 1)) {	
	  		ms.x_offset = tmp_x_offset % ms.region_range;
			tmp_reg_offset = ms.x_reg_offset + (tmp_x_offset / ms.region_range);
			if (tmp_reg_offset > (ms.side_size - 1)) 
				ms.x_reg_offset = tmp_reg_offset % ms.side_size;
			else
				ms.x_reg_offset = tmp_reg_offset;
		}
		else
			ms.x_offset = tmp_x_offset;
	}	
	else {
		tmp_x_offset = (((long) ms.x_offset) + amount);
		if ( tmp_x_offset < 0) {	
	  		ms.x_offset = tmp_x_offset % ms.region_range;
			
			tmp_reg_offset = ms.x_reg_offset + ((tmp_x_offset - ms.region_range) / ms.region_range);
			if (tmp_reg_offset < 0) {
				tmp_reg_offset = tmp_reg_offset % ms.side_size;
				ms.x_reg_offset = ms.side_size + tmp_reg_offset;
			}
			else
				ms.x_reg_offset = tmp_reg_offset;
		}
		else
			ms.x_offset = tmp_x_offset;
	}	
}

/*
Outputs a particular maps regions to the canvas
*/
void 
display_some_regions(regs, min_x_reg, max_x_reg)
	Region_Struct *regs;
	int min_x_reg, max_x_reg;
{
	int         x,y,i,j;
    	unsigned char *line;
    	long	values[2]; 
	Region_Struct *reg;
	long leftregbdr;
	long leftregbdr2;
	long bottomregbdr;
	Boolean drawtwice;
	

    	for(x = min_x_reg; x <= max_x_reg; x++) {
		drawtwice = FALSE;
		leftregbdr = ((x + ms.x_reg_offset) * ms.region_range) + (long) ms.x_offset;

		if (leftregbdr  > ((ms.side_size - 1) * ms.region_range)) {
			if (leftregbdr < ms.x.max_vis) {
				leftregbdr2 = leftregbdr;
				drawtwice = TRUE;
			}
			leftregbdr -= ms.max_range;
		}
		else if (leftregbdr < 0) {
			if ((leftregbdr + ms.region_range) > 0) {
				leftregbdr2 = leftregbdr;
				drawtwice = TRUE;
			}

			leftregbdr += ms.max_range;
		}

		for(y = ms.y.min_vis_reg; y <= ms.y.max_vis_reg; y++) {	
			bottomregbdr = (long) (y * ms.region_range);
			reg = &regs[(x * ms.side_size) + y];
    			for (i = 0; i < reg->nvect; ++i) {
        			line = reg->line[i];
        			bgnline();
        			for (j = 0; j < reg->nseg[i]; ++j) {
					if (j != 0)
	               				line += 2;
					values[0] = leftregbdr + ((long) (*line));
	       				values[1] = bottomregbdr + ((long) (*(line + 1))); 
					if ((ms.includes & INCL_3D) == INCL_3D)
					{
						if (convert_values(values) == True)
						{
							v2i(values);
						}
					}
					else
					{
						v2i(values);
					}
				}
        			endline(); 
    			}
		}

		if (drawtwice) {
			for(y = ms.y.min_vis_reg; y <= ms.y.max_vis_reg; y++) {	
				reg = &regs[(x * ms.side_size) + y];
    				for (i = 0; i < reg->nvect; ++i) {
        				line = reg->line[i];
        				bgnline();
        				for (j = 0; j < reg->nseg[i]; ++j) {
						if (j != 0)
               						line += 2;
						values[0] = leftregbdr2 + (*line);
	       					values[1] = (long) (((int) (*(line + 1))) + (y * ms.region_range)); 
						if ((ms.includes & INCL_3D) == INCL_3D)
						{
							if (convert_values(values) == True)
							{
								v2i(values);
							}
						}
						else
						{
							v2i(values);
						}
					}
        				endline(); 
				}
			}
    		}	
	}
}


void
display_regions(regs) 
	Region_Struct *regs;
{
	if (ms.x.min_vis_reg <= ms.x.max_vis_reg) 
	{

		display_some_regions(regs, ms.x.min_vis_reg, ms.x.max_vis_reg);
	}
	else 
	{
		display_some_regions(regs, ms.x.min_vis_reg, (ms.side_size- 1));
		display_some_regions(regs, 0, ms.x.max_vis_reg);
	}
}	


/* zoom undo function */
/* infinite number of undo's of zoom (restricted by memory only) */
Boolean
UndoMap(flag)
int flag;
{
	static pzoom_undo_stack base, top;

	switch (flag) {
		case ZOOM_SAVE:
			if (base == NULL) {
				base = (pzoom_undo_stack) malloc(sizeof(zoom_undo_stack));
				base->next = NULL;
				base->prev = NULL;
				base->map_state = ms;
				top = base;
				return True;
			}
			top->next = (pzoom_undo_stack)malloc(sizeof(zoom_undo_stack));
			top->next->map_state = ms;
			top->next->prev = top;
			top = top->next;
			break;
		case ZOOM_UNDO:
			if (base == NULL)
				return False;
			if (base->next == NULL) {
				ms = base->map_state;
				base = NULL;
				top = base;
				return True;
			}
			ms = top->map_state;
			top = top->prev;
			top->next = NULL;
			break;
		case ZOOM_CLEAR:
			if (base != NULL) {
				top->next = NULL;
				top->prev = NULL;
				top = NULL;
				base->next = NULL;
				base->prev = NULL;
				base = NULL;
			}
			break;
	}
	return True;
}


Boolean convert_values(values)
long *values;
{
	double alpha, beta;
	long tmp;

	if ((values[0] > 256) && (values[0] < 738))
	{
		values[0]-=512;
		values[1]-=512;

		alpha = (double)values[0]/(double)1024 * (double)6.28;
		beta = (double)values[1]/(double)512.0 * (double)1.57;

		values[0] = 384.0 * cos(beta) * sin(alpha) + 512.0;
		values[1] = 384.0 * sin(beta) + 512.0;
		return True;
	}
	else
	{
		return False;
	}
}

