#include "geotraceman.h"
#include "trace_resolve.h"

#define SUCCESS 0

/*
 * The following struct defines the data we're storing and searching
 * on. We keep an array of these.
 */
struct entry_t
{
	unsigned int ip[4];		/* IP number of this entry */
	unsigned char size;		/* number of significant bytes */
	struct entry_t *next;		/* next entry */
	char *netname;			/* textual name of entry */
	float easting,northing;		/* the geographic data */
};
static struct entry_t table[256];	/* table of networks */

float dms_read ();
static int entry_add ();
static struct entry_t *entry_find ();
int table_loadhosts ();
int table_loadnets ();
void resolve_host ();
void ParseGPOS ();

/* mikes code replaced by craigs
float dms_read (fp)
FILE *fp;
{
	unsigned int tmp[3];
	char direction[4];
	float result = 0.0;

	tmp[0] = tmp[1] = tmp[2] = 0;

	switch (fscanf(fp,"%d %d %d%c",
		&tmp[0],&tmp[1],&tmp[2],&direction[0]))
	{
		case 4: result += tmp[2] / 3600.0;
		case 3: result += tmp[1] / 60.0;
		case 2: result += tmp[0] / 1.0;
			break;
		default:
			perror("dms_read");
			break;
	}
	
	if (tolower(direction[0]) == 's' || tolower(direction[0]) == 'w')
		result = -result;

	return result;
}
*/

float dms_read (fp)
FILE *fp;
{
        unsigned int tmp[3];
        char direction;
        float result = 0.0;
	char dot;

        tmp[0] = tmp[1] = tmp[2] = 0;

	/* read in the degrees and minutes of one entry, seconds get read later */
        fscanf(fp,"%d %d%c", &tmp[0], &tmp[1], &direction);
	result+=tmp[1]/60.0;
	result+=tmp[0];

	if( tolower(direction) == 's' || tolower(direction) == 'w')
        {
                result = -result;
                return result;
        }

        if( tolower(direction) == 'n' || tolower(direction) == 'e')
        {
                return result;
        }

	/* if seconds exist, read them now */
        fscanf(fp,"%d%c", &tmp[2], &direction);
        result += tmp[2] / 3600.0;
        if( tolower(direction) == 's' || tolower(direction) == 'w')
        {
                result = -result;
                return result;
        }
                return result;
}

	
/*
 * This function adds a new IP number and location to the table.
 */
static int entry_add (ipstring,netname,northing,easting)
char *ipstring,*netname;
float northing,easting;
{
	struct entry_t *entry,*prev,*current;
	int cmp, sig;
        char *x;

	/*
	 * Allocate some space for our new entry.
	 */
	entry = (struct entry_t *)malloc(sizeof(*entry));
	if (entry == NULL)
	{
		perror("malloc");
		return 0;
	}

	/*
	 * Put some useful information into the entry.
	 */
	memset(entry->ip,0,sizeof(entry->ip));

        sig=0;
        x=ipstring;
        while( *x++ != '\0')
              if ( *x == '.')
                  sig++;

	if(sig+1 == 1)
            sscanf(ipstring,"%d",&entry->ip[0]);
	if(sig+1 == 2)
            sscanf(ipstring,"%d.%d", &entry->ip[0],&entry->ip[1]);
	if(sig+1 == 3)
            sscanf(ipstring,"%d.%d.%d", &entry->ip[0],&entry->ip[1],
                        &entry->ip[2]);
	if(sig+1 == 4)
            sscanf(ipstring,"%d.%d.%d.%d", &entry->ip[0],
			&entry->ip[1], &entry->ip[2], &entry->ip[3]);
	      
        entry->size = sig + 1;

/*
	entry->size = sscanf(ipstring,"%d.%d.%d.%d",
		&entry->ip[0],&entry->ip[1],&entry->ip[2],&entry->ip[3]);
*/
	entry->next = NULL;
	entry->netname = strdup(netname);
	entry->northing = northing;
	entry->easting = easting;

	/*
	 * Loop through the entries until we find the right place to
	 * insert the new entry.
	 */
	prev = &table[entry->ip[0]];
	current = prev->next;
	while (current != NULL)
	{
		/*
		 * If the precision of the current entry is lower than
		 * the new one, insert the new one.
		 */
		if (current->size < entry->size)
			break;

		/*
		 * Otherwise, if the precision of the current entry is equal
		 * to the new one, we have to compare the IP addresses.
		 */
		if (current->size == entry->size)
		{
			cmp = memcmp(current->ip,entry->ip,sizeof(current->ip[0]) * current->size);

			/*
			 * The two addresses are equal - this is a Bad Thing.
			 */
			if (cmp == 0)
			{
				fprintf(stderr,"entry_add: duplicate entry\n");
				free(entry);
				return 0;
			}

			/*
			 * If second is lower, insert the new entry here.
			 */
			if (cmp > 0)
				break;
		}

		/*
		 * Otherwise, we skip to the next entry and try again.
		 */
		prev = current;
		current = current->next;
	}

	/*
	 * We've found the place to add the new entry - do it.
	 */
	prev->next = entry;
	entry->next = current;
	return 1;
}

/*
 * This function searches the table for the specified IP address, and
 * returns the best matching entry or NULL.
 */
static struct entry_t *entry_find (ipstring)
char *ipstring;
{
	struct entry_t entry,*current;
        int sig;
        char *x;

	/*
	 * Put some useful information into the entry.
	 */
	memset(entry.ip,0,sizeof(entry.ip));

        sig=0;
        x=ipstring;
        while( *x++ != '\0')
              if ( *x == '.')
                  sig++;
        entry.size = sig;

        if(sig+1 == 1)
            sscanf(ipstring,"%d",&entry.ip[0]);
        if(sig+1 == 2)
            sscanf(ipstring,"%d.%d", &entry.ip[0],&entry.ip[1]);
        if(sig+1 == 3)
            sscanf(ipstring,"%d.%d.%d", &entry.ip[0],&entry.ip[1],
                        &entry.ip[2]);
        if(sig+1 == 4)
            sscanf(ipstring,"%d.%d.%d.%d", &entry.ip[0],
                        &entry.ip[1], &entry.ip[2], &entry.ip[3]);

/*
	entry.size = sscanf(ipstring,"%d.%d.%d.%d",
		&entry.ip[0],&entry.ip[1],&entry.ip[2],&entry.ip[3]);
*/
	entry.next = NULL;

	current = table[entry.ip[0]].next;
	while (current != NULL)
	{
		/*
		 * If we match to the precision of the current entry,
		 * exit and return that entry - done.
		 */
		if (!memcmp(entry.ip,current->ip,sizeof(entry.ip[0]) * current->size))
			break;

		/*
		 * Otherwise, we continue on with the next entry.
		 */
		current = current->next; 
        }
	return current;
}
	
int table_loadhosts (fname)
char *fname;
{
	FILE *fp;
	float easting,northing;
	char ipstring[256],netname[256];

	/*
	 * Open the file.
	 */
	fp = fopen(fname,"r");
	if (fp == NULL)
	{
		perror("table_load");
		return 0;
	}

	while (!feof(fp))
	{
		fscanf(fp,"%s %s",ipstring,netname);
		northing = dms_read(fp);
		easting = dms_read(fp);
		entry_add(ipstring,netname,northing,easting);
	}

	fclose(fp);
}

int table_loadnets (fname)
char *fname;
{
	FILE *fp;
	float easting,northing;
	char ipstring[256],netname[256];

	/*
	 * Open the file.
	 */
	fp = fopen(fname,"r");
	if (fp == NULL)
	{
		perror("table_load");
		return 0;
	}

	while (!feof(fp))
	{
		fscanf(fp,"%s",ipstring);
		northing = dms_read(fp);
		easting = dms_read(fp);
		fscanf(fp,"%s",netname);
		entry_add(ipstring,netname,northing,easting);
	}

	fclose(fp);
}

/*
 * This is the main interface function - resolve an IP number to a
 * network name and geographic location.
 */
void resolve_host (host, sig_size)
Host_Struct *host;
char *sig_size;
{
	struct entry_t *entry;
	int result;
        struct hostent *host_ent;
        unsigned long ad;


	
	/* GPOS query goes in here if option is set.... */

	if ((entry = entry_find(host->ipnum)) != NULL)
	{
		if (entry->size == 4)
			strcpy(host->ipname,entry->netname);
		else
		{
                    ad = inet_addr(host->ipnum);
                    if ((host_ent = gethostbyaddr((char *) &ad, 4, AF_INET)) == NULL)
				strcpy(host->ipname,entry->netname);
                    else
                                strcat(host->ipname, host_ent->h_name);

		}
		
		host->northing = entry->northing * 3600.0;
		host->easting = entry->easting * 3600.0;
		host->resolved = 1;
                *sig_size = entry->size;
	}
	else {
		strcpy(host->ipname,host->ipnum);
		*sig_size=0;
	}
}

