/*
 * Functions that deal with the repainting of the canvas widget, and the determining of
 *  the new map state after configurations
 */

#include "geotraceman.h"
#include "items.h"

/*
 * Prototypes
 */

void update_display();
Boolean draw_city();
extern Boolean convert_values();
Boolean contin_rot();

/*
Translates from pixel coordinates to mapping coordinates
*/
void
pixcoord_to_mapcoord(pixx, pixy, mapx, mapy)
	int pixx;
	int pixy;
	long *mapx;
	long *mapy;
{
	float tmpx;
	float tmpy;

	tmpx = (((float) pixx) / ((float) screen_width)) * ((float) (ms.x.max_vis - ms.x.min_vis)) + (float) ms.x.min_vis;
	tmpy = (((float) (screen_height - pixy)) / ((float) screen_height)) * ((float) (ms.y.max_vis - ms.y.min_vis)) + (float) ms.y.min_vis;

	*mapx = (long) tmpx;
	*mapy = (long) tmpy;
}

void
determine_zoom(nms, boxdata)
	Map_State_Struct *nms;
	Bounding_Box_Struct boxdata;
{
	long x1, y1, x2, y2;
	float x_zoom, y_zoom;
	long values[2];

	if ((ms.includes & INCL_3D) == INCL_3D)
	{
		values[0] = boxdata.startx;
		values[1] = boxdata.starty;
		convert_values(values);
		boxdata.startx = values[0];
		boxdata.starty = values[1];

		values[0] = boxdata.finalx;
		values[1] = boxdata.finaly;
		convert_values(values);
		boxdata.finalx = values[0];
		boxdata.finaly = values[1];
	}
	/* Convert the pixel coordinates into map coordinates */
	/* start is the upper left corner, finish is the lower right */
	pixcoord_to_mapcoord(boxdata.startx, boxdata.starty, &x1, &y1);
	pixcoord_to_mapcoord(boxdata.finalx, boxdata.finaly, &x2, &y2);

	if (x1 < x2) {
		nms->x.min_vis = x1;
		nms->x.max_vis = x2;
	}
	else {
		nms->x.min_vis = x2;
		nms->x.max_vis = x1;
	}		

	if (y1 < y2) {
		nms->y.min_vis = y1;
		nms->y.max_vis = y2;
	}
	else {
		nms->y.min_vis = y2;
		nms->y.max_vis = y1;
	}		

	/* Calculate what the new zoom level is */
	x_zoom = (float) (ms.x.max_vis - ms.x.min_vis);
	x_zoom = x_zoom / ((float) (nms->x.max_vis - nms->x.min_vis));	

	y_zoom = (float) (ms.y.max_vis - ms.y.min_vis);
	y_zoom = y_zoom / ((float) (nms->y.max_vis - nms->y.min_vis));	

	nms->zoom = x_zoom * y_zoom * ms.zoom;
}

/*
Determine the map detail from the current zoom level
*/		
int
determine_detail(nms)
	Map_State_Struct *nms;
{
	Boolean correct = False;
	int det = 0;

	while((correct == False) && (det < Map_Levels_Detail)) {
		if (nms->zoom <= (Map_Swap_Levels[det])) 
			correct = True;
		else
			det++;
	}

	if (correct == True) {
		nms->detail = det;
		return(0);
	}
	else {
		if (nms->zoom <= Map_Swap_Levels[Map_Levels_Detail]) {
			nms->detail = Map_Levels_Detail - 1;
			return(0);
		}
		else
			return(1);
	}
}

/*
Determine the minimum and maximum visible regions,
taking into account the rotation, for the x delimiters.
*/
void
update_region_delimiters()
{
	int tmp_reg;
	long virtual_min_vis, virtual_max_vis;

	virtual_min_vis = ms.x.min_vis - (ms.x_reg_offset * ms.region_range) - ms.x_offset;
	virtual_max_vis = ms.x.max_vis - (ms.x_reg_offset * ms.region_range) - ms.x_offset;
	if (virtual_min_vis < 0)
		virtual_min_vis = virtual_min_vis + ms.max_range;
	if (virtual_max_vis < 0)
		virtual_max_vis = virtual_max_vis + ms.max_range;

	ms.x.min_vis_reg = virtual_min_vis / ms.region_range;
	ms.x.max_vis_reg = virtual_max_vis / ms.region_range;

	if (ms.x.min_vis_reg == ms.x.max_vis_reg) {
		if (virtual_max_vis < virtual_min_vis) {
			ms.x.max_vis_reg--;
			if (ms.x.max_vis_reg < 0)
				ms.x.max_vis_reg = ms.x.max_vis_reg + ms.side_size;
		}
	}	

	ms.y.min_vis_reg = (ms.y.min_vis / ms.region_range);
	ms.y.max_vis_reg = (ms.y.max_vis / ms.region_range);

/*fprintf(stderr, "xmin %d, xmax %d\n", ms.x.min_vis_reg, ms.x.max_vis_reg);*/
/*fprintf(stderr, "ymin %d, ymax %d\n", ms.y.min_vis_reg, ms.y.max_vis_reg);*/
/*fprintf(stderr,"%d New %d:%d, %d:%d\n", ms.max_range, ms.x.min_vis, ms.x.max_vis, ms.y.min_vis, ms.y.max_vis);*/
}

/*
After a swap of maps, update the map status struct
*/
void
update_map_state(nms)
	Map_State_Struct *nms;
{
	long tmp; /* To hold all sorts of temporary numbers */

	ms.zoom = nms->zoom;

	/* Determine the new minimum and maximum visible limits */
	ms.x.min_vis = (long) (((float)(nms->x.min_vis * ms.max_range)) / ((float)nms->max_range)); 
	ms.x.max_vis = (long) (((float)(nms->x.max_vis * ms.max_range)) / ((float)nms->max_range)); 
	ms.y.min_vis = (long) (((float)(nms->y.min_vis * ms.max_range)) / ((float)nms->max_range)); 
	ms.y.max_vis = (long) (((float)(nms->y.max_vis * ms.max_range)) / ((float)nms->max_range)); 

/*fprintf(stderr,"%d Old %d:%d, %d:%d\n", nms->max_range, nms->x.min_vis, nms->x.max_vis, nms->y.min_vis, nms->y.max_vis);*/
/*fprintf(stderr,"%d New %d:%d, %d:%d\n", ms.max_range, ms.x.min_vis, ms.x.max_vis, ms.y.min_vis, ms.y.max_vis);*/

	/* Determine the new x offset and x region offset */
	tmp = (long) (((float)((nms->x_reg_offset * nms->region_range) + nms->x_offset)) *
			(((float) (ms.max_range)) / ((float) (nms->max_range))));	
	ms.x_reg_offset = tmp / ms.region_range;
	ms.x_offset = tmp % ms.region_range;

/*fprintf(stderr, "New X region offset %d, new x offset %d\n", ms.x_reg_offset, ms.x_offset);*/
	/* Update the region delimiters */
	update_region_delimiters();
}

/*
Process the last zoom, determining whether the map detail level should be changed, or if the zoom
has bypassed some maximum level. If the detail level is changed, then the maps are swapped
*/
int /* Returns zero if successful, or one if the zoom is not possible */ 
process_zoom(boxdata)
	Bounding_Box_Struct boxdata;
{
	Map_State_Struct nms;
	float fontx, fonty;
	Widget tmpWidget;
	long x1,x2,y1,y2;

	/* Copy the details that will be changed to the new map state */

	pixcoord_to_mapcoord(boxdata.startx, boxdata.starty, &x1, &y1);
	pixcoord_to_mapcoord(boxdata.finalx, boxdata.finaly, &x2, &y2);
	if ((x1 < ms.x.min_vis) || (x2 > ms.x.max_vis) || (y1 < ms.y.min_vis) || (y2 > ms.y.max_vis))
		return 1;
	nms.detail = ms.detail;
	nms.x_reg_offset = ms.x_reg_offset;
	nms.x_offset = ms.x_offset;
	nms.zoom = ms.zoom;

	nms.x.min_vis = ms.x.min_vis;
	nms.x.max_vis = ms.x.max_vis;

	nms.y.min_vis = ms.y.min_vis;
	nms.y.max_vis = ms.y.max_vis;

	nms.max_range = ms.max_range;
	nms.region_range = ms.region_range;
	nms.side_size = ms.side_size;

	/* Determine the new zoom level from these new coordinates */
	determine_zoom(&nms, boxdata);

	/* Determine the new map detail level, if it is too high, then give an error message
	   and keep the state as it is. Otherwise if the map state is different than current
	   then swap the maps and update the state. If the detail level is the same, then
	   simply return */

	if (determine_detail(&nms) == 1) {
		DisplayMessage(toplevel, "Error", "Zoom level not possible");
		return(1);
	}
	else {
		if (nms.detail != ms.detail) {
			ms.detail = nms.detail;
			swap_maps();
			update_map_state(&nms);
		}
		else {
			ms.zoom = nms.zoom;
			ms.x.min_vis = nms.x.min_vis;
			ms.x.max_vis = nms.x.max_vis;
			ms.y.min_vis = nms.y.min_vis;
			ms.y.max_vis = nms.y.max_vis;
			update_region_delimiters();
		}
	}

	font_size_x(&fontx);
	font_size_y(&fonty);
	htextsize(fontx, fonty);

	return(0);
}	

Boolean
read_city( fp, city)
FILE *fp;
struct city_struct *city;
{
	int c;

	/* read in the geographic location for this city */
	if (fscanf(fp,"%s", city->name) == EOF)
		return(False);

	city->northing = dms_read(fp);
	city->easting = dms_read(fp);
	/* strip off remaining chars to CR */
	for( c = fgetc( fp) ; c != '\n' ; c = fgetc( fp) );

	return(True);
}

/*
Displays all of the cities in the cities file to the screen
*/
void
cities_display()
{
	FILE *cf;
	float coord[2];
	struct city_struct city;
	int zoom_level;
	float xfont, yfont;
	int i;
	char filename[80];
	pcity_list head, next, prev;

	if ((ms.includes & INCL_CIT) != INCL_CIT)
		return;

	head = next = prev = NULL;

	if ((int)ms.zoom > ZOOMLEVEL3)
		zoom_level = 3;
	else if ((int)ms.zoom > ZOOMLEVEL2)
		zoom_level = 2;
	else 
		zoom_level = 1;

	font_size_x(&xfont);
	font_size_y(&yfont);
	htextsize(xfont, yfont);

	color(GREEN);

	for (i=1;i<zoom_level+1;i++)
	{
		sprintf(filename, "cities.file.%d", i);
		if ((cf = fopen(filename, "r")) == NULL)
			continue;

		while (read_city( cf, &city))
		{
			/* write a city name to the map */
			if (draw_city(&city, coord, head, zoom_level) == False)
				continue;
			hfont("futura.l");
			move2(coord[0],coord[1]);
			/*color(Map_Color_Cit);*/
			hcharstr(city.name);

			/* initialise linked list */
			if (head == NULL) {
				head = (pcity_list)malloc(sizeof(city_list));
				head->next = NULL;
				head->city = city;
			}
			/* move city into list */
			else {
				next = head;
				while (next != NULL) {
					prev = next;
					next = next->next;
				}
				next = (pcity_list)malloc(sizeof(city_list));
				next->city = city ;
				next->next = NULL;
				prev->next = next;
			}
		}
	fclose(cf);
	}
	swapbuffers();
}


Boolean
draw_city(city, coord, head, zoom_level)
struct city_struct *city;
float coord[2];
pcity_list head;
int zoom_level;
{
	pcity_list next;
	float coord_tmp[2];
	long values[2];

	/* Transform the coordinates */
	if (city->ndirect == 's')
		city->northing = -(city->northing);
	
	if (city->edirect == 'w')
		city->easting = -(city->easting);
	city->northing = (3600.0 * city->northing);
	city->easting = (3600.0 * city->easting);
	/* convert real world coordinates to map coords */

	if (!get_host_display_coord(city->easting, city->northing, coord))
		return False;


	if ((coord[0] < ms.x.min_vis) || (coord[0] > ms.x.max_vis))
		return False;
	if ((coord[1] < ms.y.min_vis) || (coord[1] > ms.y.max_vis))
		return False;


	if ((ms.includes & INCL_3D) == INCL_3D)
	{
		if ( (sqr(coord[0] - 512) + sqr(coord[1] - 512) ) > sqr(384) )
			return False;	
	}

	/* test for "closeness" of cities.  If too close, return False */
	/* different levels of zoom require different handling */
	next = head;
	switch (zoom_level) {
		case 1:
		    return True;
		    break;
		case 2:
		    while (next != NULL) {
		    	get_host_display_coord(next->city.easting, next->city.northing, coord_tmp);
			if ((coord[0]+20 > coord_tmp[0]) && (coord[0]-20 < coord_tmp[0]))
			    return False;
			if ((coord[1]+5 > coord_tmp[1]) && (coord[1]-5 < coord_tmp[1]))
			    return False;
			next = next->next;
		    }
	    	    break;
		case 3:
		    while (next != NULL) {
		    	get_host_display_coord(next->city.easting, next->city.northing, coord_tmp);
			if ((coord[0]+35 > coord_tmp[0]) && (coord[0]-35 < coord_tmp[0]))
			    return False;
			if ((coord[1]+25 > coord_tmp[1]) && (coord[1]-25 < coord_tmp[1]))
			    return False;
			next = next->next;
		    }
		    break;
	}
	return True;
}


/*
Called when there has been a config, or when there has been a change in the map viewing position
ie a zoom or a rotate.
*/
void
update_display()
{
	Widget tmpWidget;
	int zoom_level;
	char tmp_str[80];

	if (((ms.includes & INCL_3D) == INCL_3D) && (ms.zoom > 1.0))
	{
		ms.includes &= ~INCL_3D;
		unsetTick(ITEM_INCLUDE_3D);
	}

	/* Initialise the canvas */
	color(Map_Color_BG);
	clear();
	color(Map_Color_FG);

	/* Set the orthogonality of the canvas */
	ortho2((float) ms.x.min_vis, (float) ms.x.max_vis, (float) ms.y.min_vis, (float) ms.y.max_vis);

	/* Display the appropriate coastlines map */
	if ((ms.includes & INCL_3D) == INCL_3D)
	{
		color(Map_Color_FG);
        	circleprecision(128);
        	circ(512.0, 512.0, 384.0);
        	circleprecision(32);
	}
	display_regions(cil_map);	

	/* Test to see if river maps and political boundary maps are to be displayed, and the cities cache */

	if ((ms.includes & INCL_PDY) == INCL_PDY) {
		color(Map_Color_Pby);
		display_regions(pdy_map);
	}

	if ((ms.includes & INCL_RIV) == INCL_RIV) {
		color(Map_Color_Riv);
		display_regions(riv_map);
	}

	if ((ms.includes & INCL_CIT) == INCL_CIT) {
		color(Map_Color_Cit);
		cities_display();
	}

	swapbuffers();

	/* Display the traces */
	display_traces();

	swapbuffers();
	if ((tmpWidget = XtNameToWidget(toplevel, "*mstatusnLabel")) == NULL)
		fail0("No such widget: *mstatusnLabel");
	setLabel(tmpWidget, "waiting");
	if ((tmpWidget = XtNameToWidget(toplevel, "*zoomnLabel")) == NULL)
       	      	fail0("No such widget zoomnLabel");

	if ((int)ms.zoom >= ZOOMLEVEL3)
		zoom_level = 3;
	else if ((int)ms.zoom >= ZOOMLEVEL2)
		zoom_level = 2;
	else
		zoom_level = 1;

	sprintf(tmp_str, "%d", zoom_level);
       	setLabel(tmpWidget , tmp_str);
        if ((tmpWidget = XtNameToWidget( toplevel, "*currentnzoomLabel")) == NULL)
                fail0("No such widget currentnzoomLabel");
	sprintf(tmp_str, "%d:1", (int)ms.zoom);
        setLabel(tmpWidget , tmp_str); 
}	


void Contin_Rotate(widget, client_data, call_data)
Widget widget;
XtPointer client_data, call_data;
{
	Widget dialog = (Widget)client_data;
	char *direction = XawDialogGetValueString(dialog);

	DestroyPopupPrompt(NULL, (XtPointer)client_data, (XtPointer)NULL);

	if (tolower(direction[0]) == 'r')
	{
		ms.rot_dir = 'r';
		ms.rotating = True;
		XtAppAddWorkProc(appContext, (XtWorkProc) contin_rot, (caddr_t)NULL);
	}
	else if (tolower(direction[0]) == 'l')
	{
		ms.rot_dir = 'l';
		ms.rotating = True;
		XtAppAddWorkProc(appContext, (XtWorkProc) contin_rot, (caddr_t)NULL);
	}
	else
	{
		ms.rot_dir = '?';
		ms.rotating = False;
	}
}


Boolean contin_rot()
{
	if (ms.rotating == True)
	{
		if (ms.rot_dir == 'r')
		{
	                update_x_offset(36);
			update_region_delimiters();
			update_display();
			swapbuffers();
		}
		else if (ms.rot_dir == 'l')
		{
			update_x_offset(-36);
			update_region_delimiters();
			update_display();
			swapbuffers();
		}
		return False;
	}
	else
	{
		return True;
	}
}

